
class RequestData:

    def __init__(self, event):
        json_data = event.json()
        self.data = json_data
        self.input = json_data.get("input")
        self.shots = json_data.get("shots")
        self.device_id = json_data.get("deviceId")
        self.sdk = json_data.get("sdk").lower()
        self.server_url = json_data.get("serverUrl")



from json import JSONDecodeError

import requests

from .dataUtils import RequestData
from .utilities import JobResponse


class Backend:
    def __init__(self, request_data: RequestData, circuit):
        self.server_url = request_data.server_url
        self.sdk = request_data.sdk
        self.input = request_data.input
        self.shots = request_data.shots
        self.required_qubit = self.get_qubit_number(circuit)
        self.backend_request = self.generate_backend_request(request_data.device_id)
        # Get backend data from server
        self.backend_data = self.get_backend_data()

    def get_qubit_number(self, circuit) -> int:

        if self.sdk == "qiskit":
            return int(circuit.num_qubits)
        if self.sdk == "braket":
            return int(circuit.qubit_count)
        if self.sdk == "cirq":
            return int(len(circuit.all_qubits()))
        return 0
        # if self.sdk == "qsharp":
        # return int(self.input)  # Temporarily

    def generate_backend_request(self, device_id):

        backend_request = {
            "deviceId": device_id,
            "qubitAmount": self.required_qubit
        }

        return backend_request

    def get_backend_data(self):
        response = requests.get(
            self.server_url,
            params=self.backend_request
        )
        if response.status_code == 200:
            try:
                backend_data = response.json().get("data")
            except JSONDecodeError:
                return None
            return backend_data
        else:
            return None

    def submit_job(self, circuit, interval: int = 3, timeout: int = 30) -> JobResponse:
        job_response = JobResponse()
        shots = self.shots
        if circuit and self.backend_data and 0 < interval <= timeout:
            if self.sdk == "qiskit":
                from .sdk.qiskit import QiskitFaaS

                be_instance = QiskitFaaS(self.backend_data)
                job_response = be_instance.submit_job(circuit, shots=shots)
            elif self.sdk == "braket":
                from .sdk.braket import BraketFaaS

                be_instance = BraketFaaS(self.backend_data)
                job_response = be_instance.submit_job(circuit, shots=shots)
            elif self.sdk == "cirq":
                from .sdk.cirq import CirqFaaS

                be_instance = CirqFaaS(self.backend_data)
                job_response = be_instance.submit_job(circuit, shots=shots)
            elif self.sdk == "q#":
                from .sdk.qsharp import QsharpFaaS

                be_instance = QsharpFaaS(self.backend_data)
                job_response = be_instance.submit_job(circuit, self.input)
        elif self.backend_data is None:
            job_response.job_status = "ERROR"
        return job_response






class JobResponse(object):
    def __init__(
            self,
            provider_job_id: str = "",
            job_status: str = "",
            job_result=None,
    ):
        self.provider_job_id = provider_job_id if provider_job_id else ""
        self.job_status = job_status if job_status else "ERROR"
        self.job_result = job_result if job_result else None


class Utils:
    def __init__(self):
        # do nothing
        pass

    @staticmethod
    def generate_response(job_response: JobResponse) -> dict:
        if job_response:
            status_code = 201  # not yet finished
            if job_response.job_status == "DONE":
                status_code = 200
            job_dict = {
                "jobStatus": job_response.job_status,
                "providerJobId": job_response.provider_job_id,
                "jobResult": job_response.job_result
            }  # Object to directory
            response = {"statusCode": status_code, "body": job_dict}
        else:
            response = {
                "statusCode": 500,
                "body": "Error in function code. Please contact the developer.",
            }
        return response

    @staticmethod
    def qrng_counts_post_process(job) -> JobResponse:
        # If input type = JobResponse or have jobRawResult
        job_raw_result = {}
        if type(job) is JobResponse:
            if job.job_result:
                job_raw_result = job.job_result
        else:
            job = JobResponse()
            return job

        if job_raw_result:
            job.job_result = PostProcess.handle_job_result(job_raw_result)
            job.job_status = "DONE"
        return job


# Post-Processing functions
class PostProcess:
    def __init__(self):
        # do nothing
        pass

    @staticmethod
    def handle_job_result(job_counts: dict) -> dict:
        key_maximum_result = max(job_counts, key=job_counts.get)
        value_maximum_result = max(job_counts.values())
        all_possible_values = {
            k: int(k, 2) for k, v in job_counts.items() if v == value_maximum_result
        }
        details = {
            "decimalValue": int(key_maximum_result, 2),
            "numberOfOccurrence": value_maximum_result,
            "allPossibleValues": all_possible_values,
        }
        return {"data": key_maximum_result, "details": details}



from qiskit import IBMQ, transpile, Aer

from ..utilities import JobResponse


class QiskitFaaS:

    def __init__(self, backend_data):
        self.token = backend_data.get("token")
        self.device_name = backend_data.get("deviceName")
        # self.hub = backend_data.get("hub")
        # self.device_type = backend_data.get("deviceType")
        self.hub = None
        self.device_type = 'QUANTUM_SIMULATOR'
        self.backend = self.get_qiskit_backend()

    def init_ibmq_provider(self):
        if IBMQ.active_account():
            if IBMQ.active_account().get("token") == self.token:
                provider = IBMQ.get_provider(hub=self.hub)
            else:
                IBMQ.disable_account()
                provider = IBMQ.enable_account(token=self.token, hub=self.hub)
        else:
            provider = IBMQ.enable_account(self.token, hub=self.hub)
        return provider

    def get_qiskit_backend(self):
        if self.device_type == "QUANTUM_MACHINE":
            ibmq = self.init_ibmq_provider()
            return ibmq.get_backend(self.device_name)
        else:
            return Aer.get_backend(self.device_name)

    def transpile_circuit(self, circuit):
        return transpile(circuit, self.backend)

    def run_job(self, circuit, shots):
        # Transpile circuit to adapt with backend basis gate
        qc = self.transpile_circuit(circuit)
        job = self.backend.run(qc, shots=shots)
        return job

    def submit_job(self, qcircuit, shots) -> JobResponse:
        if self.device_type == "QUANTUM_SIMULATOR":
            try:
                job = self.run_job(qcircuit, shots)
                job_result = dict(job.result().get_counts())
                job_status = "DONE"
            except Exception:
                job_result = {}
                job_status = "ERROR"
            return JobResponse(
                provider_job_id="Internal-Qiskit-Simulation-Job",
                job_status=job_status,
                job_result=job_result,
            )
        return JobResponse()
